"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.backtraceTerm = void 0;
const enums_1 = require("../proto/enums");
const globals_1 = require("../query-builder/globals");
const has_implicit_var_1 = require("../query-builder/has-implicit-var");
const query_config_1 = require("../query-builder/query-config");
const util_1 = require("../util");
function nextBacktrace(i, backtrace) {
    if (backtrace && backtrace[0] === i) {
        return backtrace.slice(1);
    }
}
function joinMultiArray(acc, next) {
    return acc[0]
        ? [`${acc[0]}, ${next[0]}`, `${acc[1]}  ${next[1]}`]
        : [next[0], next[1]];
}
function getMarked(str, backtrace) {
    const s = Array.isArray(str) ? str[0] : str;
    const emptyMarks = Array.isArray(str) ? str[1] : ' '.repeat(str.length);
    return backtrace && backtrace.length === 0
        ? [s, '^'.repeat(s.length)]
        : [s, emptyMarks];
}
function combineMarks(literals, ...placeholders) {
    let result = '';
    let mark = '';
    for (let i = 0; i < placeholders.length; i++) {
        result += literals[i];
        mark += ' '.repeat(literals[i].length);
        if (!Array.isArray(placeholders[i])) {
            result += placeholders[i];
            mark += ' '.repeat(placeholders[i].length);
        }
        else {
            result += placeholders[i][0];
            mark += placeholders[i][1];
        }
    }
    // add the last literal
    result += literals[literals.length - 1];
    mark += ' '.repeat(literals[literals.length - 1].length);
    return [result, mark];
}
function backtraceObject(obj, backtrace) {
    const [param, ...nextB] = backtrace || [];
    if (obj.$reql_type$ === 'BINARY') {
        return getMarked('<Buffer>', backtrace);
    }
    if (Object.keys(obj).length === 0) {
        return getMarked('{}', backtrace);
    }
    return combineMarks `{ ${Object.entries(obj)
        .map(([key, val]) => {
        const next = param === key ? nextB : undefined;
        return getMarked(combineMarks `${(0, util_1.snakeToCamel)(key)}: ${backtraceTerm(val, false, next)}`);
    })
        .reduce(joinMultiArray, ['', ''])} }`;
}
function backtraceTerm(term, head = true, backtrace) {
    const parseArg = (arg, index, all, forceHead = false) => backtraceTerm(arg, forceHead || (!all && index === 0), nextBacktrace(index, backtrace));
    if (term === undefined) {
        return getMarked('');
    }
    if (!Array.isArray(term)) {
        let termStr = ['', ''];
        if (term === null) {
            termStr = getMarked('null');
        }
        else if (typeof term === 'object') {
            termStr = backtraceObject(term, backtrace);
        }
        else if (typeof term === 'string') {
            termStr = getMarked(`"${term}"`);
        }
        else {
            termStr = getMarked(term.toString());
        }
        return getMarked(head ? combineMarks `r.expr(${termStr})` : termStr, backtrace);
    }
    const [type, args, optarg] = term;
    const hasArgs = !!args && !!args.length;
    switch (type) {
        case enums_1.TermType.MAKE_ARRAY: {
            if (!args) {
                return getMarked('');
            }
            return getMarked(head
                ? combineMarks `r.expr([${args
                    .map(parseArg)
                    .reduce(joinMultiArray, ['', ''])}])`
                : combineMarks `[${args
                    .map(parseArg)
                    .reduce(joinMultiArray, ['', ''])}]`, backtrace);
        }
        case enums_1.TermType.IMPLICIT_VAR:
            return getMarked('r.row', backtrace);
        case enums_1.TermType.FUNC: {
            if (!!args && Array.isArray(args) && (0, has_implicit_var_1.hasImplicitVar)(term)) {
                return backtraceTerm(args[1], false, nextBacktrace(1, backtrace));
            }
            const paramsBacktrace = nextBacktrace(0, backtrace);
            const params = args[0][1].map((i) => getMarked(`var_${i}`, nextBacktrace(i, paramsBacktrace)));
            const funcHead = params.reduce(joinMultiArray, ['', '']);
            const body = backtraceTerm(args[1], false, nextBacktrace(1, backtrace));
            return getMarked(globals_1.globals.backtraceType === 'lambda'
                ? combineMarks `(${funcHead}) => ${body}`
                : combineMarks `function(${funcHead}) { return ${body} }`, backtrace);
        }
        case enums_1.TermType.VAR: {
            return getMarked(`var_${args[0]}`, backtrace);
        }
        case enums_1.TermType.FUNCALL: {
            if (!args) {
                return getMarked('');
            }
            const [func, caller, ...params] = args;
            const parsedParams = params
                .map((a, i) => parseArg(a, i + 2))
                .reduce(joinMultiArray, ['', '']);
            const parsedFunc = parseArg(func, 0);
            const parsedCaller = parseArg(caller, 1, undefined, true);
            return getMarked(parsedParams[0]
                ? combineMarks `${parsedCaller}.do(${parsedParams}, ${parsedFunc})`
                : combineMarks `${parsedCaller}.do(${parsedFunc})`, backtrace);
        }
        case enums_1.TermType.BRACKET: {
            if (!args) {
                return getMarked('');
            }
            const [caller, ...params] = args;
            const parsedParams = [...params]
                .map((a, i) => parseArg(a, i + 1))
                .reduce(joinMultiArray, ['', '']);
            return getMarked(combineMarks `${parseArg(caller, 0)}(${parsedParams})`, backtrace);
        }
        default: {
            const c = query_config_1.rConsts.find((co) => co[0] === type);
            if (c) {
                return getMarked(`r.${c[1]}`, backtrace);
            }
            const func = query_config_1.termConfig.find((conf) => conf[0] === type);
            if (!func) {
                const rfunc = query_config_1.rConfig.find((conf) => conf[0] === type);
                if (rfunc) {
                    const rparsedParams = [...(args || [])]
                        .map(parseArg)
                        .reduce(joinMultiArray, ['', '']);
                    if (optarg) {
                        const wrappedOptarg = backtraceObject(optarg, backtrace);
                        return getMarked(hasArgs
                            ? combineMarks `r.${rfunc[1]}(${rparsedParams}, ${wrappedOptarg})`
                            : combineMarks `r.${rfunc[1]}(${wrappedOptarg})`, backtrace);
                    }
                    return getMarked(combineMarks `r.${rfunc[1]}(${rparsedParams})`, backtrace);
                }
                return getMarked('');
            }
            if (!args) {
                return getMarked(combineMarks `r.${func[1]}(${backtraceObject(optarg, backtrace)})`, backtrace);
            }
            const [caller, ...params] = args;
            // TODO: Maybe this is better but should fix tests
            // if (
            //   typeof caller === 'object' &&
            //   (!Array.isArray(caller) || caller[0] === TermType.MAKE_ARRAY)
            // ) {
            //   const rparsedParams = [...(args || [])]
            //     .map(parseArg)
            //     .reduce(joinMultiArray, ['', '']);
            //   return getMarked(
            //     optarg
            //       ? hasArgs
            //         ? combineMarks`r.${func[1]}(${rparsedParams}, ${backtraceObject(
            //             optarg,
            //             backtrace
            //           )})`
            //         : combineMarks`r.${func[1]}(${backtraceObject(
            //             optarg,
            //             backtrace
            //           )})`
            //       : combineMarks`r.${func[1]}(${rparsedParams})`,
            //     backtrace
            //   );
            // }
            const hasParams = params.length > 0;
            const parsedParams = params
                .map((a, i) => parseArg(a, i + 1))
                .reduce(joinMultiArray, ['', '']);
            const parsedCaller = parseArg(caller, 0);
            const parsedOptarg = optarg
                ? backtraceObject(optarg, backtrace)
                : undefined;
            if (parsedOptarg) {
                return getMarked(hasParams
                    ? combineMarks `${parsedCaller}.${func[1]}(${parsedParams}, ${parsedOptarg})`
                    : combineMarks `${parsedCaller}.${func[1]}(${parsedOptarg})`, backtrace);
            }
            return getMarked(combineMarks `${parsedCaller}.${func[1]}(${parsedParams})`, backtrace);
        }
    }
}
exports.backtraceTerm = backtraceTerm;
